# -*- coding: utf-8 -*-
"""cubicweb-ctl plugin providing the index-in-es command

:organization: Logilab
:copyright: 2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""
from __future__ import print_function

import os.path as osp

from elasticsearch.helpers import parallel_bulk

from cubicweb.cwctl import CWCTL
from cubicweb.utils import admincnx
from cubicweb.toolsutils import Command

from cubes.elasticsearch.es import indexable_types, fulltext_indexable_rql


__docformat__ = "restructuredtext en"

HERE = osp.dirname(osp.abspath(__file__))


class IndexInES(Command):

    """Index content in ElasticSearch.

    <instance id>
      identifier of the instance

    """
    name = 'index-in-es'
    min_args = max_args = 1
    arguments = '<instance id>'
    options = [('dry-run', {'type': 'yn', 'default': False,
                            'help': 'set to True if you want to skip the insertion in ES'}),
               ('debug', {'type': 'yn', 'default': False,
                          'help': 'set to True if you want to print'
                                  'out debug info and progress'}),
               ('etypes', {'type': 'csv', 'default': '',
                           'help': 'only index given etypes '
                                   '[default:all indexable types]'}),
               ('index-name', {'type': 'string', 'default': '',
                               'help': 'override index-name if you want to use a different ID'
                                       '[default: uses index-name from all-in-one.conf]'}),
               ('except-etypes', {'type': 'string', 'default': '',
                                  'help': 'all indexable types except given etypes'
                                          '[default: []]'}),
               ]

    def run(self, args):
        """run the command with its specific arguments"""
        appid = args.pop(0)
        with admincnx(appid) as cnx:
            schema = cnx.vreg.schema
            indexer = cnx.vreg['es'].select('indexer', cnx)
            es = indexer.get_connection()
            indexer.create_index()
            if self.config.index_name:
                indexer.create_index(index_name=self.config.index_name)
            if es:
                if self.config.etypes:
                    etypes = self.config.etypes
                else:
                    etypes = indexable_types(schema,
                                             custom_skip_list=self.config.except_etypes)
                    assert self.config.except_etypes not in etypes
                if self.config.debug and not self.config.etypes:
                    print(u'found indexable_types {}'.format(
                        ','.join(etypes)))
                for _ in parallel_bulk(es,
                                       self.bulk_actions(etypes,
                                                         cnx,
                                                         index_name=self.config.index_name,
                                                         dry_run=self.config.dry_run),
                                       raise_on_error=False,
                                       raise_on_exception=False):
                            pass
            else:
                if self.config.debug:
                    print(u'no elasticsearch configuration found, skipping')

    def bulk_actions(self, etypes, cnx,
                     index_name=None, dry_run=False):
        for etype in etypes:
            rql = fulltext_indexable_rql(etype, cnx.vreg.schema)
            rset = cnx.execute(rql)
            if self.config.debug:
                print(u'indexing {} {}'.format(etype, len(rset)))
                print(u'RQL : {}'.format(rql))
            for entity in rset.entities():
                serializer = entity.cw_adapt_to('IFullTextIndexSerializable')
                json = serializer.serialize()
                if not dry_run:
                    data = {'_op_type': 'index',
                            '_index': index_name or cnx.vreg.config['index-name'],
                            '_type': etype,
                            '_id': entity.eid,
                            '_source': json
                            }
                    self.customize_data(data)
                    yield data

    def customize_data(self, data):
        pass


CWCTL.register(IndexInES)



import unittest
import time

from elasticsearch_dsl import Search

from cubicweb.devtools import testlib

from cubes.elasticsearch.testutils import RealESTestMixin, BlogFTIAdapter
from cubes.elasticsearch.search_helpers import compose_search


class ReindexOnRelationTests(RealESTestMixin, testlib.CubicWebTC):

    def test_es_hooks_modify_relation(self):
        with self.admin_access.cnx() as cnx:
            with self.temporary_appobjects(BlogFTIAdapter):
                indexer = cnx.vreg['es'].select('indexer', cnx)
                indexer.create_index(custom_settings={
                    'mappings': {
                        'BlogEntry': {'_parent': {"type": "Blog"}},
                    }
                })
                blog1 = cnx.create_entity('Blog', title=u'Blog')
                entity = cnx.create_entity('BlogEntry',
                                           title=u'Article about stuff',
                                           content=u'yippee',
                                           entry_of=blog1)
                blog2 = cnx.create_entity('Blog', title=u'Blog')
                cnx.commit()
                time.sleep(2)  # TODO find a way to have synchronous operations in unittests
                search = compose_search(Search(index=self.config['index-name'],
                                               doc_type='Blog'),
                                        'yippee',
                                        parents_for="BlogEntry",
                                        fields=['_all'])
                results = search.execute()
                self.assertEquals(len(results), 1)
                self.assertCountEqual([hit.eid for hit in results],
                                      [blog1.eid])
                blog2.cw_set(reverse_entry_of=entity)
                cnx.commit()
                time.sleep(2)  # TODO find a way to have synchronous operations in unittests
                search = compose_search(Search(index=self.config['index-name'],
                                               doc_type='Blog'),
                                        'yippee',
                                        parents_for="BlogEntry",
                                        fields=['_all'])
                results = search.execute()
                self.assertEquals(len(results), 2)
                self.assertCountEqual([hit.eid for hit in results],
                                      [blog1.eid, blog2.eid])


if __name__ == '__main__':
    unittest.main()



from __future__ import print_function

import time
import unittest

from elasticsearch_dsl import Search

from cubicweb.devtools import testlib

from cubes.elasticsearch.search_helpers import compose_search

from cubes.elasticsearch.testutils import RealESTestMixin, BlogFTIAdapter


class ParentsSearchTC(RealESTestMixin, testlib.CubicWebTC):

    def test_parent_search(self):
        with self.admin_access.cnx() as cnx:
            with self.temporary_appobjects(BlogFTIAdapter):
                indexer = cnx.vreg['es'].select('indexer', cnx)
                indexer.create_index(custom_settings={
                    'mappings': {
                        'BlogEntry': {'_parent': {"type": "Blog"}},
                    }
                })
                test_structure = {
                    u'A': [u'Paris ceci', u'Nantes', u'Toulouse'],
                    u'B': [u'Paris cela'],
                    u'C': [u'Paris autre', u'Paris plage'],
                }
                for fa_title, facomp_contents in test_structure.items():
                    blog = cnx.create_entity('Blog',
                                             title=fa_title)
                    for facomp_content in facomp_contents:
                        cnx.create_entity('BlogEntry',
                                          entry_of=blog,
                                          title=facomp_content,
                                          content=facomp_content)
                cnx.commit()
            time.sleep(2)  # TODO find a way to have synchronous operations in unittests
            for query, number_of_results, first_result in (("Paris", 3, "C"),
                                                           ("Nantes", 1, "A")):
                search = compose_search(Search(index=self.config['index-name'],
                                               doc_type='Blog'),
                                        query,
                                        parents_for="BlogEntry",
                                        fields=['_all'],
                                        fuzzy=True)
                self.assertEquals(len(search.execute()), number_of_results)
                self.assertEquals(search.execute().to_dict()['hits']['hits'][0]['_source']['title'],
                                  first_result)


if __name__ == '__main__':
    unittest.main()
